#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <sys/time.h>

#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
//#include <dlfnc.h>

#include "prolog.h"
//#include "cdi.h"	// interface to user program
#include "tinecdi.h"	// interface to user program
#include "client-lib.h"
#include "cdiLiberaLib.h"

#define WAIT_TIME_NS 1000000  // 1ms
int nLiberaLines = N_LIBERA_LINES;

__attribute__((constructor)) void cdiLiberaMain(void)
{
    int cc = 0;

    if ((cc=cdiRegisterBusFilter("LIBERA",filterLiberaBus)) != 0 ) goto err;
    if ((cc=cdiRegisterBusInitialization("LIBERA",initLiberaForCDI)) != 0 ) goto err;
    if ((cc=cdiRegisterBusHandler("LIBERA",execLiberaBus)) != 0 ) goto err;
    if ((cc=cdiRegisterBusCleanup("LIBERA",exitLibera)) != 0 ) goto err;
err:
    if (cc)
    {
      cdilog("%s : failure in bus registration LIBERA",erlst[cc]);
    }
    return;
}

LIBERA_STRUCT liberaInfo[N_LIBERA_LINES] = {
	{-1, -1}, {-1, -1}, {-1, -1}, {-1, -1}
};

void printLiberaError( int err, const char *errstr )
{
	if ( -1 == err ) {

		perror( errstr );
	}
	else {
 		fprintf( stderr, "%s: %s", errstr, cspi_strerror(err) );
 		if ( CSPI_E_SYSTEM == err ) {
 			fprintf( stderr, ": %s", strerror(errno) );
		}
	
		fputc('\n', stderr);
	}
}

/* ============  init routine  =================================== */
int initLibera(int ln, CdiRequestInfoBlk *pReq)
{ char fname[80];
  LIBERA_STRUCT *pLibera;
  int rc = 0;

  printf("Libera input devParameters: %d;%d; == %d; %d; %d, %d, %d, %d, %d\n",pReq->pDev->cdiAddr, pReq->pDev->cdiSubAddr, pReq->devParameters[0],pReq->devParameters[1],pReq->devParameters[2],pReq->devParameters[3],pReq->devParameters[4],pReq->devParameters[5],pReq->devParameters[6]);

  sprintf(liberaInfo[ln].ipadr, "%d.%d.%d.%d", pReq->devParameters[liberaIP],
      pReq->devParameters[liberaIP + 1],pReq->devParameters[liberaIP + 2],pReq->devParameters[liberaIP + 3]);

  if ( (rc = server_connect(liberaInfo[ln].ipadr, 0, 0, 0 )) != 0 )
  {
    printLiberaError( rc, "server_connect" );
    cdilog("Open Libera line, server_connect Error! line: %d\n",ln+1);
    return -not_ready;
  }
 	if ( CSPI_OK != (rc = cspi_allochandle( CSPI_HANDLE_ENV, 0, &liberaInfo[ln].cspiEnvHen )) )
  {
    printLiberaError( rc, "cspi_allochandle" );
    cdilog("Open Libera line, cspi_allochandle Error! line: %d\n",ln+1);
    return -not_ready;
  }
  pLibera = &liberaInfo[ln];
  pLibera->LiberaLine = ln;

  cdilog("Open Libera line OK! line number: %d\n",ln+1);	
  return 0;
}

// ============  exit routine  ===================================
int exitLibera(int ln)
{ LIBERA_STRUCT *pLibera;
  int rc;

  if (ln<1 || ln>nLiberaLines) return -illegal_line;
  ln--;

  pLibera = &liberaInfo[ln];
  if ( pLibera->LiberaLine < 0 ) return -file_not_open;

  rc = server_disconnect();
	if ( CSPI_OK != rc )
  {
    printLiberaError( rc, "server_disconnect" );
    cdilog("Close Libera line, server_disconnect Error! line: %d\n",ln+1);
  }  
	if ( CSPI_OK == rc )
	  if ( CSPI_OK != (rc = cspi_freehandle( CSPI_HANDLE_ENV, pLibera->cspiEnvHen )) )
    {
      printLiberaError( rc, "cspi_freehandle" );
      cdilog("Close Libera line, cspi_freehandle Error! line: %d\n",ln+1);
    }
	if ( CSPI_OK == rc ) cdilog("close Libera line OK! line number: %d", ln+1);

  pLibera->LiberaLine = -1;

  return 0;
}

int filterLiberaBus(int *parameter, int accessFlag)
{
  return 0;
}

int initLiberaForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{ int ret = 0;
  
  if (busLine<1 || busLine>N_LIBERA_LINES) ret = -illegal_line; 
//  ret = initLibera(busLine);
  if ( ret >= 0 ) liberaInfo[busLine - 1].cdiLine = cdiLine;

  return ret;
}

int doLiberaRequest(CdiRequestInfoBlk *pReq, int step)
{ int access, ln, ret, length, writeOffset, done;
  void *pOut;

  if ( step == 1 ) /* first request */
    access = (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag && pReq->accessFlag != cdiReadReadFlag) ? 1 : 0;
  else if ( step == 2 )/* second */
    access = (pReq->accessFlag == cdiReadWriteFlag || pReq->accessFlag == cdiWriteWriteFlag ) ? 1 : 0;
  else
    access = pReq->accessFlag == cdiWriteReadReadFlag ? 0 : 1;

  ln = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine - 1;

  if ( liberaInfo[ln].LiberaLine < 0 ) 
    if ( ret = initLibera(ln, pReq) )
    {
      pReq->errorCode = -ret;
      if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      pReq->pDev->numberDeviceError++;
      pReq->active = finishRequest; /* finish request */
      return pReq->errorCode;
    }
/*
  ret = canMsgSend((int)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine,
              (short)(pReq->devParameters[canBroadcast] ? 0 : pReq->pDev->cdiAddr),
              (short)pReq->devParameters[access ? canOpenId : canOpenReadId], 
              (short)pReq->outLength, 
              (char *)pReq->pOutData,
              (short)pReq->devParameters[canOpenObj], 
              (short)pReq->devParameters[canOpenSubIndex], 
              (short)access);
*/
  if ( ret != 0 )
  {
    pReq->errorCode = ret;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    if ( ret > 0 )
    {
      pReq->deviceError = ret;    // return from hardware
      pReq->pDev->numberDeviceError++;
    }
    pReq->active = finishRequest; /* finish request */
  }

  return pReq->errorCode;
}

/********************************************************/
// if sync call, data not pass filter then repeat Transmit
int checkData(CdiRequestInfoBlk *pReq)
{
  struct timespec tv;

  /* check return condition */
  checkReturnCondition(pReq, (char *)pReq->bufLastData, (char *)pReq->pInData, pReq->inLength, 0);
  if ( pReq->dataReturnCondition == DATA_PASS_FILTER)
    memcpy( pReq->bufLastData, pReq->pInData, pReq->inLength < 8 ? pReq->inLength : 8);

 	if ( pReq->conditionFlag > anyData && pReq->dataReturnCondition != DATA_PASS_FILTER )
	{
    if ( pReq->callback == NULL )
    {
      if ( getTimeMilliseconds() - pReq->requestTime >= pReq->timeout )
      {
         pReq->errorCode = invalid_data;	// data back for sync call
         if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      }
      else
      {
        tv.tv_sec = 0; tv.tv_nsec = WAIT_TIME_NS; // delay 1ms
        nanosleep( &tv, NULL );    
        return 1;
      }
    }
    pReq->active = finishRequest;
  }
  return 0;
}
/********************************************************/
/* only call once with pReq->active == transRequest */
/********************************************************/
void execLiberaBus(CdiRequestInfoBlk *pReq)
{ 
  if ( pReq->accessFlag > cdiWriteReadFlag )
  {
    pReq->errorCode = not_implemented;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    pReq->active = finishRequest; /* finish request */
    return;
  }

  /* do first part; */
  do
  {
    doLiberaRequest(pReq, 1);
    if ( pReq->active == finishRequest ) return;
    /* check return condition */
    if (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if ( pReq->active == finishRequest ) return;
  if ( pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiWriteFlag)
  {
    pReq->active = finishRequest;
    return;
  }

  /* do second part */
  do
  {
    doLiberaRequest(pReq, 2);
    if ( pReq->active == finishRequest ) return;
    /* check return condition */
    if (pReq->accessFlag != cdiWriteReadFlag && pReq->accessFlag != cdiWriteReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if (pReq->accessFlag != cdiWriteReadWriteFlag && pReq->accessFlag != cdiWriteReadReadFlag)
  {
    pReq->active = finishRequest; /* finish request */
    return;
  }

  /* do third part */
  doLiberaRequest(pReq, 3);
  pReq->active = finishRequest; /* finish request */
}

int msleep(int msec)
{
  struct timespec ts;
  if (msec <= 0) return 0;
  ts.tv_sec = msec/1000;
  ts.tv_nsec = (msec%1000) * 1000000;
  return nanosleep(&ts,NULL);
}
