#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cdi.h"
#include "tinecdi.h"
#include "tinetype.h"

#if defined(UNIX)
CDI_EXPORT int cdiCreateThread(void *cdiThreadParam, int index)
{
  pthread_t tid;

  if (pthread_create(&tid,NULL,index == THREAD_INDEX_MASTER ? ThreadMaster : ThreadCallback,cdiThreadParam) != 0)
  {
    perror("pthread_create");
    cdilog("CDI: cannot create thread!");
    return 0;
  }
  return (int)tid;
}
CDI_EXPORT int cdiCancelThread(int cdiThreadId)
{
  return pthread_cancel(cdiThreadId);
}
CDI_EXPORT int cdiGetSemaphore(sem_t *sem,int access,int count)
{
  int cc = 0;
  if (sem == NULL) return -1;
  if ((cc=sem_init(sem, access, count)) != 0)
  {
    perror("sem_init");
    cdilog("CDI: cannot initialize semaphore");
  }
  return cc;
}
CDI_EXPORT int cdiDestroySemaphore(sem_t *sem)
{
  return sem_destroy(sem);
}
CDI_EXPORT int cdiPostSemaphore(sem_t *sem)
{
  return sem_post(sem);
}
CDI_EXPORT int cdiWaitSemaphore(sem_t *sem,int peek)
{
  if (peek) return sem_trywait(sem);
  return sem_wait(sem);
}
CDI_EXPORT int cdiMutexInit(cdi_mutex_t *mutex, int line)
{
  return pthread_mutex_init(mutex,NULL);
}
CDI_EXPORT int cdiMutexLock(cdi_mutex_t *mutex)
{
  return pthread_mutex_lock(mutex);
}
CDI_EXPORT int cdiMutexUnlock(cdi_mutex_t *mutex)
{
  return pthread_mutex_unlock(mutex);
}
CDI_EXPORT int cdi_nanosleep(const struct timespec *req, struct timespec *rem)
{
  return nanosleep(req,rem);
}
#elif defined(WIN32)
//#define KEEP_OLDWAY
CDI_EXPORT int cdi_nanosleep(const struct timespec *req, struct timespec *rem)
{
# ifndef KEEP_OLDWAY
  return nanosleep(req,rem);
# else
  int msec;
  if (req == NULL) return -1;
  msec = req->tv_sec*1000 + req->tv_nsec/1000000;
  Sleep(msec);
  return 0;
# endif
}
CDI_EXPORT int cdiCreateThread(void *cdiThreadParam, int index)
{
  DWORD dwCdiThreadId;
  return (int)CreateThread(NULL,0,index == THREAD_INDEX_MASTER ? ThreadMaster : ThreadCallback,cdiThreadParam,0,&dwCdiThreadId);
}
CDI_EXPORT int cdiCancelThread(int cdiThreadId)
{
  if (TerminateThread((HANDLE)cdiThreadId,0) == 0) return -1;
  if (CloseHandle((HANDLE)cdiThreadId) == 0) return -2;
  return 0;
}
CDI_EXPORT int cdiGetSemaphore(sem_t *sem,int access,int count)
{
  HANDLE hSem = CreateSemaphore(NULL,count,CDI_SEM_MAXCOUNT,NULL);
  if (sem != NULL) *sem = hSem;
  return hSem == NULL ? -1 : 0;
}
CDI_EXPORT int cdiDestroySemaphore(sem_t *sem)
{
  if (sem == NULL) return -1;
  return CloseHandle((HANDLE)(*sem));
}
CDI_EXPORT int cdiPostSemaphore(sem_t *sem)
{
  int rc;
  if (sem == NULL) return -1;
  rc = ReleaseSemaphore((HANDLE)*sem,1,NULL);
  return rc ? 0 : -1;
}
CDI_EXPORT int cdiWaitSemaphore(sem_t *sem,int peek)
{
  int to = peek ? 0 : INFINITE;
  if (sem == NULL) return -1;
  if (WaitForSingleObject((HANDLE)*sem,to) == WAIT_TIMEOUT) return -1;
  return 0;
}
CDI_EXPORT int cdiMutexInit(cdi_mutex_t *mutex, int id)
{ int lastError;
  char lpName[64];

  if (mutex == NULL) return -1;

  if (id < 0)
  {
    strcpy(lpName,"CdiMutexLogFile");
  }
  else
  {
    sprintf(lpName,"CdiMutex:%d",id);
  }

  if ((*mutex=CreateMutex(NULL,FALSE,lpName)) == NULL) 
  {
    lastError = GetLastError();
    cdilog("cdiMutexInit %s : 0x%x",lpName,lastError);
    return -1;
  }
  if (id >= 0)
  {
    cdilog("cdiMutexInit %s : success",lpName);
  }
  return 0;
}
CDI_EXPORT int cdiMutexLock(cdi_mutex_t *mutex)
{
  return WaitForSingleObject(*mutex,INFINITE) == WAIT_FAILED ? -1 : 0;
}
CDI_EXPORT int cdiMutexUnlock(cdi_mutex_t *mutex)
{
  return ReleaseMutex(*mutex) ? 0 : -1;
}
#elif defined(VXWORKS)
CDI_EXPORT int cdiCreateThread(void *cdiThreadParam, int index)
{
  /* taskSpawn() ? */
  return 0;
}
CDI_EXPORT int cdiMutexLock(cdi_mutex_t *mutex)
{
  return 0;
}
CDI_EXPORT int cdiMutexUnlock(cdi_mutex_t *mutex)
{
  return 0;
}
#endif /* UNIX, WIN32, VXWORKS */

